@@ -222,8 +222,7 @@ static void ste_dial(struct ofono_voicecall *vc,
 		return;
 
 error:
-	if (cbd)
-		g_free(cbd);
+	g_free(cbd);
 
 	CALLBACK_WITH_FAILURE(cb, data);
 }
@@ -248,8 +247,7 @@ static void ste_template(const char *cmd, struct ofono_voicecall *vc,
 		return;
 
 error:
-	if (req)
-		g_free(req);
+	g_free(req);
 
 	CALLBACK_WITH_FAILURE(cb, data);
 }
@@ -315,8 +313,7 @@ static void ste_release_specific(struct ofono_voicecall *vc, int id,
 		return;
 
 error:
-	if (req)
-		g_free(req);
+	g_free(req);
 
 	CALLBACK_WITH_FAILURE(cb, data);
 }
@@ -401,8 +398,7 @@ static void ste_send_dtmf(struct ofono_voicecall *vc, const char *dtmf,
 		return;
 
 error:
-	if (cbd)
-		g_free(cbd);
+	g_free(cbd);
 
 	CALLBACK_WITH_FAILURE(cb, data);
 }
@@ -533,6 +529,16 @@ static void ecav_notify(GAtResult *result, gpointer user_data)
 	}
 }
 
+static void ste_voicecall_initialized(gboolean ok, GAtResult *result,
+					gpointer user_data)
+{
+	struct ofono_voicecall *vc = user_data;
+	struct voicecall_data *vd = ofono_voicecall_get_data(vc);
+
+	g_at_chat_register(vd->chat, "*ECAV:", ecav_notify, FALSE, vc, NULL);
+	ofono_voicecall_register(vc);
+}
+
 static int ste_voicecall_probe(struct ofono_voicecall *vc, unsigned int vendor,
 				void *data)
 {
@@ -540,13 +546,12 @@ static int ste_voicecall_probe(struct ofono_voicecall *vc, unsigned int vendor,
 	struct voicecall_data *vd;
 
 	vd = g_new0(struct voicecall_data, 1);
-	vd->chat = chat;
+	vd->chat = g_at_chat_clone(chat);
 
 	ofono_voicecall_set_data(vc, vd);
 
-	g_at_chat_send(chat, "AT*ECAM=1", NULL, NULL, NULL, NULL);
-	g_at_chat_register(chat, "*ECAV:", ecav_notify, FALSE, vc, NULL);
-	ofono_voicecall_register(vc);
+	g_at_chat_send(vd->chat, "AT*ECAM=1", none_prefix,
+			ste_voicecall_initialized, vc, NULL);
 
 	return 0;
 }
@@ -560,6 +565,7 @@ static void ste_voicecall_remove(struct ofono_voicecall *vc)
 
 	ofono_voicecall_set_data(vc, NULL);
 
+	g_at_chat_unref(vd->chat);
 	g_free(vd);
 }
 
@@ -569,7 +575,7 @@ static struct ofono_voicecall_driver driver = {
 	.remove			= ste_voicecall_remove,
 	.dial			= ste_dial,
 	.answer			= ste_answer,
-	.hangup			= ste_hangup,
+	.hangup_active		= ste_hangup,
 	.hold_all_active	= ste_hold_all_active,
 	.release_all_held	= ste_release_all_held,
 	.set_udub		= ste_set_udub,
