@@ -56,14 +56,26 @@
 
 static const char *none_prefix[] = { NULL };
 static const char *sysinfo_prefix[] = { "^SYSINFO:", NULL };
+static const char *ussdmode_prefix[] = { "^USSDMODE:", NULL };
+static const char *cvoice_prefix[] = { "^CVOICE:", NULL };
+
+enum huawei_sim_state {
+	HUAWEI_SIM_STATE_INVALID_OR_LOCKED =	0,
+	HUAWEI_SIM_STATE_VALID =		1,
+	HUAWEI_SIM_STATE_INVALID_CS =		2,
+	HUAWEI_SIM_STATE_INVALID_PS =		3,
+	HUAWEI_SIM_STATE_INVALID_PS_AND_CS =	4,
+	HUAWEI_SIM_STATE_NOT_EXISTENT =		255
+};
 
 struct huawei_data {
 	GAtChat *modem;
 	GAtChat *pcui;
 	struct ofono_sim *sim;
-	gint sim_state;
+	enum huawei_sim_state sim_state;
 	struct ofono_gprs *gprs;
 	struct ofono_gprs_context *gc;
+	gboolean voice;
 };
 
 static int huawei_probe(struct ofono_modem *modem)
@@ -100,17 +112,62 @@ static void huawei_debug(const char *str, void *user_data)
 	ofono_info("%s%s", prefix, str);
 }
 
-static void notify_sim_state(struct ofono_modem *modem, gint sim_state)
+static void ussdmode_query_cb(gboolean ok, GAtResult *result,
+						gpointer user_data)
+{
+	struct huawei_data *data = user_data;
+	GAtResultIter iter;
+	gint ussdmode;
+
+	if (!ok)
+		return;
+
+	g_at_result_iter_init(&iter, result);
+
+	if (!g_at_result_iter_next(&iter, "^USSDMODE:"))
+		return;
+
+	if (!g_at_result_iter_next_number(&iter, &ussdmode))
+		return;
+
+	if (ussdmode == 0)
+		return;
+
+	/* set USSD mode to text mode */
+	g_at_chat_send(data->pcui, "AT^USSDMODE=0", none_prefix,
+						NULL, NULL, NULL);
+}
+
+static void ussdmode_support_cb(gboolean ok, GAtResult *result,
+						gpointer user_data)
+{
+	struct huawei_data *data = user_data;
+	GAtResultIter iter;
+
+	if (!ok)
+		return;
+
+	g_at_result_iter_init(&iter, result);
+
+	if (!g_at_result_iter_next(&iter, "^USSDMODE:"))
+		return;
+
+	/* query current USSD mode */
+	g_at_chat_send(data->pcui, "AT^USSDMODE?", ussdmode_prefix,
+					ussdmode_query_cb, data, NULL);
+}
+
+static void notify_sim_state(struct ofono_modem *modem,
+				enum huawei_sim_state sim_state)
 {
 	struct huawei_data *data = ofono_modem_get_data(modem);
 
-	if (data->sim_state == 0 && sim_state == 1) {
-		ofono_sim_inserted_notify(data->sim, TRUE);
-		data->sim_state = sim_state;
-	} else if (data->sim_state == 1 && sim_state == 0) {
+	if (sim_state == HUAWEI_SIM_STATE_NOT_EXISTENT)
 		ofono_sim_inserted_notify(data->sim, FALSE);
-		data->sim_state = sim_state;
-	}
+	else
+		ofono_sim_inserted_notify(data->sim, TRUE);
+
+	data->sim_state = sim_state;
 }
 
 static void sysinfo_cb(gboolean ok, GAtResult *result, gpointer user_data)
@@ -142,24 +199,188 @@ static void sysinfo_cb(gboolean ok, GAtResult *result, gpointer user_data)
 	if (!g_at_result_iter_next_number(&iter, &sim_state))
 		return;
 
-	notify_sim_state(modem, sim_state);
+	notify_sim_state(modem, (enum huawei_sim_state) sim_state);
 }
 
 static void simst_notify(GAtResult *result, gpointer user_data)
 {
 	struct ofono_modem *modem = user_data;
 	GAtResultIter iter;
-	int state;
+	gint sim_state;
 
 	g_at_result_iter_init(&iter, result);
 
 	if (!g_at_result_iter_next(&iter, "^SIMST:"))
 		return;
 
-	if (!g_at_result_iter_next_number(&iter, &state))
+	if (!g_at_result_iter_next_number(&iter, &sim_state))
 		return;
 
-	notify_sim_state(modem, state);
+	notify_sim_state(modem, (enum huawei_sim_state) sim_state);
+}
+
+static void orig_notify(GAtResult *result, gpointer user_data)
+{
+	GAtResultIter iter;
+	gint call_id, call_type;
+
+	g_at_result_iter_init(&iter, result);
+
+	if (!g_at_result_iter_next(&iter, "^ORIG:"))
+		return;
+
+	if (!g_at_result_iter_next_number(&iter, &call_id))
+		return;
+
+	if (!g_at_result_iter_next_number(&iter, &call_type))
+		return;
+
+	ofono_info("Call origin: id %d type %d", call_id, call_type);
+}
+
+static void conf_notify(GAtResult *result, gpointer user_data)
+{
+	GAtResultIter iter;
+	gint call_id;
+
+	g_at_result_iter_init(&iter, result);
+
+	if (!g_at_result_iter_next(&iter, "^CONF:"))
+		return;
+
+	if (!g_at_result_iter_next_number(&iter, &call_id))
+		return;
+
+	ofono_info("Call setup: id %d", call_id);
+}
+
+static void conn_notify(GAtResult *result, gpointer user_data)
+{
+	GAtResultIter iter;
+	gint call_id, call_type;
+
+	g_at_result_iter_init(&iter, result);
+
+	if (!g_at_result_iter_next(&iter, "^CONN:"))
+		return;
+
+	if (!g_at_result_iter_next_number(&iter, &call_id))
+		return;
+
+	if (!g_at_result_iter_next_number(&iter, &call_type))
+		return;
+
+	ofono_info("Call connect: id %d type %d", call_id, call_type);
+}
+
+static void cend_notify(GAtResult *result, gpointer user_data)
+{
+	GAtResultIter iter;
+	gint call_id, duration, end_status, cc_pause;
+
+	g_at_result_iter_init(&iter, result);
+
+	if (!g_at_result_iter_next(&iter, "^CEND:"))
+		return;
+
+	if (!g_at_result_iter_next_number(&iter, &call_id))
+		return;
+
+	if (!g_at_result_iter_next_number(&iter, &duration))
+		return;
+
+	if (!g_at_result_iter_next_number(&iter, &end_status))
+		return;
+
+	if (!g_at_result_iter_next_number(&iter, &cc_pause))
+		return;
+
+	ofono_info("Call end: id %d duration %ds status %d control %d",
+				call_id, duration, end_status, cc_pause);
+}
+
+
+static void cvoice_query_cb(gboolean ok, GAtResult *result,
+						gpointer user_data)
+{
+	struct ofono_modem *modem = user_data;
+	struct huawei_data *data = ofono_modem_get_data(modem);
+	GAtResultIter iter;
+	gint mode, rate, bits, period;
+
+	if (!ok)
+		goto done;
+
+	g_at_result_iter_init(&iter, result);
+
+	if (!g_at_result_iter_next(&iter, "^CVOICE:"))
+		goto done;
+
+	if (!g_at_result_iter_next_number(&iter, &mode))
+		goto done;
+
+	if (!g_at_result_iter_next_number(&iter, &rate))
+		goto done;
+
+	if (!g_at_result_iter_next_number(&iter, &bits))
+		goto done;
+
+	if (!g_at_result_iter_next_number(&iter, &period))
+		goto done;
+
+	data->voice = TRUE;
+
+	ofono_info("Voice channel: %d Hz, %d bits, %dms period",
+						rate, bits, period);
+
+	g_at_chat_register(data->pcui, "^ORIG:", orig_notify,
+						FALSE, NULL, NULL);
+	g_at_chat_register(data->pcui, "^CONF:", conf_notify,
+						FALSE, NULL, NULL);
+	g_at_chat_register(data->pcui, "^CONN:", conn_notify,
+						FALSE, NULL, NULL);
+	g_at_chat_register(data->pcui, "^CEND:", cend_notify,
+						FALSE, NULL, NULL);
+
+	/* check available voice ports */
+	g_at_chat_send(data->pcui, "AT^DDSETEX=?", none_prefix,
+						NULL, NULL, NULL);
+
+done:
+	ofono_modem_set_powered(modem, TRUE);
+
+	/* query current sim state */
+        g_at_chat_send(data->pcui, "AT^SYSINFO", sysinfo_prefix,
+						sysinfo_cb, modem, NULL);
+}
+
+static void cvoice_support_cb(gboolean ok, GAtResult *result,
+						gpointer user_data)
+{
+	struct ofono_modem *modem = user_data;
+	struct huawei_data *data = ofono_modem_get_data(modem);
+	GAtResultIter iter;
+
+	if (!ok)
+		goto done;
+
+	g_at_result_iter_init(&iter, result);
+
+	if (!g_at_result_iter_next(&iter, "^CVOICE:"))
+		goto done;
+
+	/* query current voice setting */
+	g_at_chat_send(data->pcui, "AT^CVOICE?", cvoice_prefix,
+					cvoice_query_cb, modem, NULL);
+
+	return;
+
+done:
+	ofono_modem_set_powered(modem, TRUE);
+
+	/* query current sim state */
+	g_at_chat_send(data->pcui, "AT^SYSINFO", sysinfo_prefix,
+						sysinfo_cb, modem, NULL);
 }
 
 static void cfun_enable(gboolean ok, GAtResult *result, gpointer user_data)
@@ -169,18 +390,30 @@ static void cfun_enable(gboolean ok, GAtResult *result, gpointer user_data)
 
 	DBG("");
 
-	ofono_modem_set_powered(modem, ok);
-
-	if (!ok)
+	if (!ok) {
+		ofono_modem_set_powered(modem, FALSE);
 		return;
+	}
 
 	/* follow sim state */
 	g_at_chat_register(data->pcui, "^SIMST:", simst_notify,
-							FALSE, modem, NULL);
+						FALSE, modem, NULL);
 
-	/* query current sim state */
-	g_at_chat_send(data->pcui, "AT^SYSINFO", sysinfo_prefix,
-					sysinfo_cb, modem, NULL);
+	/* query current device settings */
+	g_at_chat_send(data->pcui, "AT^U2DIAG?", none_prefix,
+						NULL, NULL, NULL);
+
+	/* query current port settings */
+	g_at_chat_send(data->pcui, "AT^GETPORTMODE", none_prefix,
+						NULL, NULL, NULL);
+
+	/* check USSD mode support */
+	g_at_chat_send(data->pcui, "AT^USSDMODE=?", ussdmode_prefix,
+					ussdmode_support_cb, data, NULL);
+
+	/* check for voice support */
+	g_at_chat_send(data->pcui, "AT^CVOICE=?", cvoice_prefix,
+					cvoice_support_cb, modem, NULL);
 }
 
 static GAtChat *create_port(const char *device)
@@ -235,7 +468,8 @@ static void huawei_disconnect(gpointer user_data)
 
 	DBG("");
 
-	ofono_gprs_context_remove(data->gc);
+	if (data->gc)
+		ofono_gprs_context_remove(data->gc);
 
 	g_at_chat_unref(data->modem);
 	data->modem = NULL;
@@ -247,13 +481,16 @@ static void huawei_disconnect(gpointer user_data)
 	g_at_chat_set_disconnect_function(data->modem,
 						huawei_disconnect, modem);
 
-	ofono_info("Reopened GPRS context channel");
+	if (data->sim_state == HUAWEI_SIM_STATE_VALID ||
+			data->sim_state == HUAWEI_SIM_STATE_INVALID_CS) {
+		ofono_info("Reopened GPRS context channel");
 
-	data->gc = ofono_gprs_context_create(modem, 0, "atmodem",
-							data->modem);
+		data->gc = ofono_gprs_context_create(modem, 0, "atmodem",
+								data->modem);
 
-	if (data->gprs && data->gc)
-		ofono_gprs_add_context(data->gprs, data->gc);
+		if (data->gprs && data->gc)
+			ofono_gprs_add_context(data->gprs, data->gc);
+	}
 }
 
 static int huawei_enable(struct ofono_modem *modem)
@@ -276,6 +513,9 @@ static int huawei_enable(struct ofono_modem *modem)
 		return -EIO;
 	}
 
+	if (ofono_modem_get_boolean(modem, "HasVoice") == TRUE)
+		data->voice = TRUE;
+
 	data->sim_state = 0;
 
 	g_at_chat_send(data->pcui, "ATE0", none_prefix, NULL, NULL, NULL);
@@ -331,10 +571,12 @@ static void huawei_pre_sim(struct ofono_modem *modem)
 	DBG("%p", modem);
 
 	ofono_devinfo_create(modem, 0, "atmodem", data->pcui);
-	data->sim = ofono_sim_create(modem, 0, "atmodem", data->pcui);
+	data->sim = ofono_sim_create(modem, OFONO_VENDOR_QUALCOMM_MSM,
+					"atmodem", data->pcui);
 
-	if (ofono_modem_get_boolean(modem, "HasVoice") == TRUE)
-		ofono_voicecall_create(modem, 0, "atmodem", data->pcui);
+	if (data->voice == TRUE)
+		ofono_voicecall_create(modem, OFONO_VENDOR_HUAWEI,
+						"atmodem", data->pcui);
 }
 
 static void huawei_post_sim(struct ofono_modem *modem)
@@ -345,26 +587,38 @@ static void huawei_post_sim(struct ofono_modem *modem)
 
 	DBG("%p", modem);
 
+	if (data->sim_state == HUAWEI_SIM_STATE_INVALID_PS_AND_CS) {
+		ofono_phonebook_create(modem, 0, "atmodem", data->pcui);
+		return;
+	}
+
 	netreg = ofono_netreg_create(modem, OFONO_VENDOR_HUAWEI, "atmodem",
 								data->pcui);
 
 	ofono_sms_create(modem, OFONO_VENDOR_HUAWEI, "atmodem", data->pcui);
-	ofono_cbs_create(modem, OFONO_VENDOR_QUALCOMM_MSM, "atmodem",
-								data->pcui);
-	ofono_ussd_create(modem, 0, "atmodem", data->pcui);
+	ofono_cbs_create(modem, OFONO_VENDOR_QUALCOMM_MSM,
+						"atmodem", data->pcui);
+	ofono_ussd_create(modem, OFONO_VENDOR_QUALCOMM_MSM,
+						"atmodem", data->pcui);
+	ofono_phonebook_create(modem, 0, "atmodem", data->pcui);
 
-	data->gprs = ofono_gprs_create(modem, 0, "atmodem", data->pcui);
-	data->gc = ofono_gprs_context_create(modem, 0, "atmodem", data->modem);
+	if (data->sim_state == HUAWEI_SIM_STATE_VALID ||
+			data->sim_state == HUAWEI_SIM_STATE_INVALID_CS) {
+		data->gprs = ofono_gprs_create(modem, 0, "atmodem", data->pcui);
+		data->gc = ofono_gprs_context_create(modem, 0, "atmodem",
+								data->modem);
 
-	if (data->gprs && data->gc)
-		ofono_gprs_add_context(data->gprs, data->gc);
+		if (data->gprs && data->gc)
+			ofono_gprs_add_context(data->gprs, data->gc);
+	}
 
-	if (ofono_modem_get_boolean(modem, "HasVoice") == TRUE) {
+	if ((data->sim_state == HUAWEI_SIM_STATE_VALID ||
+			data->sim_state == HUAWEI_SIM_STATE_INVALID_PS) &&
+							data->voice == TRUE) {
 		ofono_call_forwarding_create(modem, 0, "atmodem", data->pcui);
 		ofono_call_settings_create(modem, 0, "atmodem", data->pcui);
 		ofono_call_barring_create(modem, 0, "atmodem", data->pcui);
 		ofono_ssn_create(modem, 0, "atmodem", data->pcui);
-		ofono_phonebook_create(modem, 0, "atmodem", data->pcui);
 
 		mw = ofono_message_waiting_create(modem);
 		if (mw)
